;;; Copyright (c) 2004 Nikodemus Siivola <nikodemus@random-state.net>
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining
;;; a copy of this software and associated documentation files (the
;;; "Software"), to deal in the Software without restriction, including
;;; without limitation the rights to use, copy, modify, merge, publish,
;;; distribute, sublicense, and/or sell copies of the Software, and to
;;; permit persons to whom the Software is furnished to do so, subject to
;;; the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be
;;; included in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
;;; LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
;;; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
;;; WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :asdf-packaging-tools)

(defparameter *lint-output* t)
(defparameter *huge-form-limit* 38)

(defvar *delayed-notes*)
(defvar *context*)

;;; Lint notes

(defun prepare-file-note (pathname)
  (format nil "In ~A:~%" (namestring pathname)))

(defun format-note (string &rest args)
  (loop for note in (nreverse *delayed-notes*)
        do (format *lint-output* note))
  (setf *delayed-notes* nil)
  (apply #'format *lint-output* string args))

(defun note-huge (name size)
  (format-note "~& - ~A ~A is huge (~A lines).~%" *context* name size))

(defun note-bad-name (name string &rest args)
  (format-note "~& - Bad name: ~A~@[ in ~A~]. ~?~%" 
	       name *context* string args))

(defun note-dirty-name (name)
  (format-note "~& - ~A using interned symbol ~A. ~
                Use strings, uninterned symbols, or keywords instead.~%" 
	       *context* name))

(defun note-bad-style (info)
  (format-note "~& - ~A is considered bad style. ~A~%" *context* info))

(defun note-key-with-optional (name)
  (format-note "~& - ~A ~A has both optional and keyword arguments.~%" *context* name))

(defun note-read-eval-error (err)
  (format-note "~& - Error during READ-EVAL: ~A." err))

(defun note-aux ()
  (format-note "~& - ~A with &AUX in arglist. Use LET inside the body.~%"
	       *context*))

;;; General purpose lint routines

(defun lint-name (name-designator &optional wrap-char)
  (let ((name (string (if (consp name-designator)
			  (second name-designator)
			  name-designator))))
    (when (find #\_ name)
      (note-bad-name name "Use \"-\" as word separator."))
    (when wrap-char
      (unless (char= wrap-char (char name 0) (char name (1- (length name))))
	(note-bad-name name "Use ~C as pre- and postfix character." 
		       wrap-char)))))

(defun lint-size (name form)
  (let ((lines (count #\newline (with-output-to-string (s)
                                  (pprint form s)))))
    (when (> lines *huge-form-limit*)
      (note-huge name lines))))

(defun lint-arglist (name arglist)
  (let (key optional)
    (labels ((rec (args)
	       (when args
		 (cond ((consp args)
			  (rec (car args))
			  (rec (cdr args)))
		       ((eq '&key args)
			  (setf key t))
		       ((eq '&optional args)
		          (setf optional t))
		       ((eq '&aux args)
			  (note-aux))
		       (t
		          (lint-name args))))))
      (rec arglist))
    (when (and key optional)
      (note-key-with-optional name))))

(defun lint-clean-name (name)
  (when (or (and (symbolp name) 
		 (not (keywordp name))
		 (symbol-package name))
	    (and (consp name)
		 (starts-with 'quote name)))
    (note-dirty-name name))
  (lint-name name))

(defun lint-form (form)
  (typecase form
    (cons (lint (car form) (cdr form)))
    (symbol (lint-name form))))

(defgeneric lint (car cdr)
  (:method (car cdr)
    (lint-form car)
    (if (listp cdr)
	(dolist (form cdr)
	  (lint-form form))
	(lint-form cdr))))

;;; Special purpose lint routines

(defmacro deflint (name destructuring-arglist &body body)
  (with-unique-names (car cdr)
    `(defmethod lint ((,car (eql ',name)) ,cdr)
       (declare (ignore ,car))
       (let ((*context* ',name))
	 (destructuring-bind ,destructuring-arglist ,cdr
	   ,@body)))))

(macrolet ((def (name &optional insert)
	       `(defmethod lint ((name (eql ',name)) cdr)
		  (let ((*context* ',name))
		    (note-bad-style 
		     ,(if insert
			  (concatenate 'string
				       "Use " insert
				       ", or DESTRUCTURING bind instead.")
			  "Use DESTRUCTURING-BIND instead."))
		    (lint-form cdr)))))
  (def caar)
  (def cadr "SECOND")
  (def cdar)
  (def cddr)
  (def caaar)
  (def caadr)
  (def cadar)
  (def caddr "THIRD")
  (def cdaar)
  (def cdadr)
  (def cddar)
  (def cdddr)
  (def caaaar)
  (def caaadr)
  (def caadar)
  (def caaddr)
  (def cadaar)
  (def cadadr)
  (def caddar)
  (def cadddr "FOURTH")
  (def cdaaar)
  (def cdaadr)
  (def cdadar)
  (def cdaddr)
  (def cddaar)
  (def cddadr)
  (def cddddr))

(macrolet ((def (name)
	       `(defmethod lint ((name (eql ',name)) cdr)
		  (let ((*context* ',name))
		    (note-bad-style "Use DEFPACKAGE instead.")
		    (lint-form cdr)))))
  (def use-package)
  (def make-package)
  (def shadow)
  (def shadowing-import)
  (def import))

(deflint in-package (package)
  (lint-clean-name package))

(macrolet ((def (name)
	       `(defmethod lint ((name (eql ',name)) cdr)
		  (let ((*context* ',name))
		    (note-bad-style "Use DEFSYSTEM instead.")
		    (lint-name cdr)))))
  (def require)
  (def provide)
  (def load)
  (def compile-file))

(deflint eval (exp)
  (note-bad-style "Try using eg. MACROLET instead."))

(deflint set (symbol value)
  (note-bad-style "Use (SETF SYMBOL-VALUE) instead."))

(deflint quote (exp)
  (lint-form exp))

(deflint defclass (name &body body)
  (lint-name name)
  (lint-form body))

(deflint defconstant (name value &optional doc)
  (declare (ignore doc))
  (lint-name name #\+)
  (lint-form value))

(deflint defgeneric (name arglist &body body)
  (lint-name name)
  (lint-arglist name arglist)
  (lint-size name body)
  (lint-form body))

(deflint define-compiler-macro (name arglist &body body)
  (lint-name name)
  (lint-arglist name arglist)
  (lint-size name body)
  (lint-form body))

(deflint define-setf-expander (name arglist &body body)
  (lint-name name)
  (lint-arglist name arglist)
  (lint-size name body)
  (lint-form body))

(deflint defmacro (name arglist &body body)
  (lint-name name)
  (lint-arglist name arglist)
  (lint-size name body)
  (lint-form body))

(deflint defmethod (name &body body)
  (lint-name name)
  (multiple-value-bind (qualifier arglist body) (parse-method body)
    (lint-name qualifier)
    (lint-arglist name arglist)
    (lint-size name body)
    (lint-form body)))

(deflint defpackage (name &rest options)
  (lint-clean-name name)
  (dolist (option options)
    (destructuring-bind (name . body) option
      (unless (eq name :size)
	(dolist (name body) 
	  (lint-clean-name name))))))
	
(deflint defparameter (name value &optional doc)
  (declare (ignore doc))
  (lint-name name #\*)
  (lint-form value))

(deflint defun (name arglist &body body)
  (lint-name name)
  (lint-arglist name arglist)
  (lint-size name body)
  (lint-form body))

(deflint defvar (name &optional value doc)
  (declare (ignore doc))
  (lint-name name #\*)
  (lint-form value))

(deflint flet (functions &body body)
  (dolist (fun functions)
    (destructuring-bind (name arglist &body body) fun
      (let ((*context* `(flet ,name)))
	(lint-name name)
	(lint-arglist name arglist)
	(lint-size name body)
	(lint-form body))))
  (lint-size "body" body)
  (lint-form body))

(deflint labels (functions &body body)
  (dolist (fun functions)
    (destructuring-bind (name arglist &body body) fun
      (let ((*context* `(labels ,name)))
	(lint-name name)
	(lint-arglist name arglist)
	(lint-size name body)
	(lint-form body))))
  (lint-size "body" body)
  (lint-form body))

;;;; Hooks into ASDF

(defclass lint-op (operation)
  ((force :initform t)))

(defmethod perform ((op lint-op) x)
  nil)

(defmethod perform ((op lint-op) (src cl-source-file))
  (let ((*delayed-notes* (list (prepare-file-note 
                                (component-relative-pathname src))))
	(*readtable* (copy-readtable))
	(*context* "top-level")
	(old-eval (get-dispatch-macro-character #\# #\.)))
    (set-dispatch-macro-character #\# #\. 
				  (lambda (&rest args)
				    (multiple-value-bind (x e) 
					(ignore-errors (apply old-eval args))
				      (cond (e
					      (note-read-eval-error e)
					      (values))
					    (t x)))))
    (with-open-file (f (component-pathname src))
      (loop for form = (read f nil nil)
	    while form
	    do (lint-form form)))))
