;;; Copyright (c) 2004 Nikodemus Siivola <nikodemus@random-state.net>
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining
;;; a copy of this software and associated documentation files (the
;;; "Software"), to deal in the Software without restriction, including
;;; without limitation the rights to use, copy, modify, merge, publish,
;;; distribute, sublicense, and/or sell copies of the Software, and to
;;; permit persons to whom the Software is furnished to do so, subject to
;;; the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be
;;; included in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
;;; LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
;;; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
;;; WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :asdf-packaging-tools)

(defmacro with-unique-names ((&rest bindings) &body body)
  `(let ,(mapcar #'(lambda (binding)
		     (destructuring-bind (var prefix)
			 (if (consp binding) binding (list binding binding))
		       `(,var (gensym ,(string prefix)))))
		 bindings)
     ,@body))

(defun starts-with (exp list)
  (and (consp list) (eql exp (car list))))

(define-modify-macro appendf (&rest lists) append)

(defun parse-body (body)
  (let (doc decls)
    (loop for form = (pop body)
	  do (cond ((and (stringp form) (not doc))
		    (setf doc form))
		   ((and (consp form) (starts-with 'declare form))
		    (appendf decls (cdr form)))
		   (t 
		    (push form body)
		    (return))))
    (values doc decls body)))

(defun parse-method (qualifier-arglist-and-body)
  (values
   (unless (consp (first qualifier-arglist-and-body))
     (pop qualifier-arglist-and-body))
   (first qualifier-arglist-and-body)
   (rest qualifier-arglist-and-body)))

(defun only (list)
  (if (= 1 (length list))
      (car list)
      (error "ONLY called with ~S." list)))

(defun release-name (system)
  (concatenate 'string 
	       (component-name system) 
	       "-" 
	       (component-version system)))

(defun tarball-name (system)
  (concatenate 'string (release-name system) ".tar.gz"))

(defun signature (name)
  (concatenate 'string name ".asc"))

(defun latest-tarball-name (system)
  (concatenate 'string (component-name system) "-latest.tar.gz"))

(defun xterm-pathname ()
  ;; This is not nice. Urk.
  (loop for dir in '("/usr/bin/" "/usr/local/bin")
	do (loop for prog in '("xterm" "rxvt" "x-terminal-emulator")
		 for path = (probe-file (merge-pathnames prog dir))
		 when path
		 do (return-from xterm-pathname path))))

(defun ask (question test explanation)
  (flet ((question ()
	   (format *query-io* "~&~A~%> " question)
	   (force-output *query-io*)))
    (loop for answer = (question)
	  when (funcall test answer)
	  return answer
	  else
	  do (format *query-io* "~&~A~%" explanation))))

(defun temporary-directory (op system)
  (let ((dir (target-directory op)))
    (make-pathname :directory (append (pathname-directory dir)
				      (list (release-name system)))
		   :defaults dir)))

(define-condition shell-command-error (error)
  ((format-string :initform "" :initarg :format-string :reader format-string)
   (format-args :initform nil :initarg :format-args :reader format-args))
  (:report (lambda (error stream)
	     (format stream "Shell command failed: ~?"
		     (format-string error)
		     (format-args error)))))

(defun run-safe-shell-command (format &rest args)
  (unless (zerop (apply #'run-shell-command format args))
    (error 'shell-command-error 
	   :format-string format
	   :format-args args)))
